@@ -2,6 +2,7 @@
 
 use crate::InstructionKinds;
 use arbitrary::{Arbitrary, Result, Unstructured};
+use std::borrow::Cow;
 
 /// Configuration for a generated module.
 ///
@@ -59,6 +60,41 @@ pub trait Config: 'static + std::fmt::Debug {
         100
     }
 
+    /// The imports that may be used when generating the module.
+    ///
+    /// Defaults to `None` which means that any arbitrary import can be generated.
+    ///
+    /// To only allow specific imports, override this method to return a WebAssembly module which
+    /// describes the imports allowed.
+    ///
+    /// Note that [`Self::min_imports`] is ignored when `available_imports` are enabled.
+    ///
+    /// # Panics
+    ///
+    /// The returned value must be a valid binary encoding of a WebAssembly module. `wasm-smith`
+    /// will panic if the module cannot be parsed.
+    ///
+    /// # Example
+    ///
+    /// An implementation of this method could use the `wat` crate to provide a human-readable and
+    /// maintainable description:
+    ///
+    /// ```rust
+    /// Some(wat::parse_str(r#"
+    ///     (module
+    ///         (import "env" "ping" (func (param i32)))
+    ///         (import "env" "pong" (func (result i32)))
+    ///         (import "env" "memory" (memory 1))
+    ///         (import "env" "table" (table 1))
+    ///         (import "env" "tag" (tag (param i32)))
+    ///     )
+    /// "#))
+    /// # ;
+    /// ```
+    fn available_imports(&self) -> Option<Cow<'_, [u8]>> {
+        None
+    }
+
     /// The minimum number of functions to generate. Defaults to 0.  This
     /// includes imported functions.
     fn min_funcs(&self) -> usize {
@@ -237,40 +273,64 @@ pub trait Config: 'static + std::fmt::Debug {
     }
 
     /// Determines whether the bulk memory proposal is enabled for generating
-    /// insructions. Defaults to `false`.
+    /// instructions.
+    ///
+    /// Defaults to `false`.
     fn bulk_memory_enabled(&self) -> bool {
         false
     }
 
     /// Determines whether the reference types proposal is enabled for
-    /// generating insructions. Defaults to `false`.
+    /// generating instructions.
+    ///
+    /// Defaults to `false`.
     fn reference_types_enabled(&self) -> bool {
         false
     }
 
     /// Determines whether the SIMD proposal is enabled for
-    /// generating insructions. Defaults to `false`.
+    /// generating instructions.
+    ///
+    /// Defaults to `false`.
     fn simd_enabled(&self) -> bool {
         false
     }
 
     /// Determines whether the Relaxed SIMD proposal is enabled for
-    /// generating insructions. Defaults to `false`.
+    /// generating instructions.
+    ///
+    /// Defaults to `false`.
     fn relaxed_simd_enabled(&self) -> bool {
         false
     }
 
     /// Determines whether the exception-handling proposal is enabled for
-    /// generating insructions. Defaults to `false`.
+    /// generating instructions.
+    ///
+    /// Defaults to `false`.
     fn exceptions_enabled(&self) -> bool {
         false
     }
 
-    /// Determines whether the module linking proposal is enabled.
+    /// Determines whether the multi-value results are enabled.
     ///
-    /// Defaults to `false`.
-    fn module_linking_enabled(&self) -> bool {
-        false
+    /// Defaults to `true`.
+    fn multi_value_enabled(&self) -> bool {
+        true
+    }
+
+    /// Determines whether the nontrapping-float-to-int-conversions propsal is enabled.
+    ///
+    /// Defaults to `true`.
+    fn saturating_float_to_int_enabled(&self) -> bool {
+        true
+    }
+
+    /// Determines whether the sign-extension-ops propsal is enabled.
+    ///
+    /// Defaults to `true`.
+    fn sign_extension_ops_enabled(&self) -> bool {
+        true
     }
 
     /// Determines whether a `start` export may be included. Defaults to `true`.
@@ -335,6 +395,13 @@ pub trait Config: 'static + std::fmt::Debug {
     fn allowed_instructions(&self) -> InstructionKinds {
         InstructionKinds::all()
     }
+
+    /// Returns whether we should generate custom sections or not.
+    ///
+    /// This is false by default.
+    fn generate_custom_sections(&self) -> bool {
+        false
+    }
 }
 
 /// The default configuration.
@@ -358,51 +425,50 @@ impl Config for DefaultConfig {}
 #[derive(Clone, Debug)]
 #[allow(missing_docs)]
 pub struct SwarmConfig {
-    // These fields are configured via `Arbitrary`
-    pub max_types: usize,
-    pub max_imports: usize,
-    pub max_tags: usize,
-    pub max_funcs: usize,
-    pub max_globals: usize,
-    pub max_exports: usize,
+    pub allow_start_export: bool,
+    pub available_imports: Option<Vec<u8>>,
+    pub bulk_memory_enabled: bool,
+    pub canonicalize_nans: bool,
+    pub exceptions_enabled: bool,
+    pub max_aliases: usize,
+    pub max_data_segments: usize,
     pub max_element_segments: usize,
     pub max_elements: usize,
-    pub max_data_segments: usize,
+    pub max_exports: usize,
+    pub max_funcs: usize,
+    pub max_globals: usize,
+    pub max_imports: usize,
+    pub max_instances: usize,
     pub max_instructions: usize,
     pub max_memories: usize,
-    pub min_uleb_size: u8,
-    pub max_tables: usize,
     pub max_memory_pages: u64,
-    pub bulk_memory_enabled: bool,
-    pub reference_types_enabled: bool,
-    pub module_linking_enabled: bool,
-    pub max_aliases: usize,
+    pub max_modules: usize,
     pub max_nesting_depth: usize,
-
-    // These fields are always set to their default value as specified in the
-    // default trait implementation.
+    pub max_tables: usize,
+    pub max_tags: usize,
+    pub max_type_size: u32,
+    pub max_types: usize,
     pub memory64_enabled: bool,
-    pub min_types: usize,
-    pub min_imports: usize,
-    pub min_tags: usize,
-    pub min_funcs: usize,
-    pub min_globals: usize,
-    pub min_exports: usize,
+    pub memory_max_size_required: bool,
+    pub memory_offset_choices: (u32, u32, u32),
     pub min_data_segments: usize,
     pub min_element_segments: usize,
     pub min_elements: usize,
+    pub min_exports: usize,
+    pub min_funcs: usize,
+    pub min_globals: usize,
+    pub min_imports: usize,
     pub min_memories: u32,
     pub min_tables: u32,
-    pub max_instances: usize,
-    pub max_modules: usize,
-    pub memory_offset_choices: (u32, u32, u32),
-    pub memory_max_size_required: bool,
-    pub simd_enabled: bool,
+    pub min_tags: usize,
+    pub min_types: usize,
+    pub min_uleb_size: u8,
+    pub multi_value_enabled: bool,
+    pub reference_types_enabled: bool,
     pub relaxed_simd_enabled: bool,
-    pub exceptions_enabled: bool,
-    pub allow_start_export: bool,
-    pub max_type_size: u32,
-    pub canonicalize_nans: bool,
+    pub saturating_float_to_int_enabled: bool,
+    pub sign_extension_enabled: bool,
+    pub simd_enabled: bool,
 }
 
 impl<'a> Arbitrary<'a> for SwarmConfig {
@@ -430,8 +496,11 @@ impl<'a> Arbitrary<'a> for SwarmConfig {
             bulk_memory_enabled: reference_types_enabled || u.arbitrary()?,
             reference_types_enabled,
             simd_enabled: u.arbitrary()?,
+            multi_value_enabled: u.arbitrary()?,
             max_aliases: u.int_in_range(0..=MAX_MAXIMUM)?,
             max_nesting_depth: u.int_in_range(0..=10)?,
+            saturating_float_to_int_enabled: u.arbitrary()?,
+            sign_extension_enabled: u.arbitrary()?,
 
             // These fields, unlike the ones above, are less useful to set.
             // They either make weird inputs or are for features not widely
@@ -456,8 +525,8 @@ impl<'a> Arbitrary<'a> for SwarmConfig {
             exceptions_enabled: false,
             memory64_enabled: false,
             max_type_size: 1000,
-            module_linking_enabled: false,
             canonicalize_nans: false,
+            available_imports: None,
         })
     }
 }
@@ -479,6 +548,12 @@ impl Config for SwarmConfig {
         self.max_imports
     }
 
+    fn available_imports(&self) -> Option<Cow<'_, [u8]>> {
+        self.available_imports
+            .as_ref()
+            .map(|is| Cow::Borrowed(&is[..]))
+    }
+
     fn min_funcs(&self) -> usize {
         self.min_funcs
     }
@@ -583,10 +658,6 @@ impl Config for SwarmConfig {
         self.reference_types_enabled
     }
 
-    fn module_linking_enabled(&self) -> bool {
-        self.module_linking_enabled
-    }
-
     fn simd_enabled(&self) -> bool {
         self.simd_enabled
     }
@@ -599,6 +670,18 @@ impl Config for SwarmConfig {
         self.exceptions_enabled
     }
 
+    fn multi_value_enabled(&self) -> bool {
+        self.multi_value_enabled
+    }
+
+    fn saturating_float_to_int_enabled(&self) -> bool {
+        self.saturating_float_to_int_enabled
+    }
+
+    fn sign_extension_ops_enabled(&self) -> bool {
+        self.sign_extension_enabled
+    }
+
     fn allow_start_export(&self) -> bool {
         self.allow_start_export
     }
@@ -611,6 +694,10 @@ impl Config for SwarmConfig {
         self.max_nesting_depth
     }
 
+    fn max_type_size(&self) -> u32 {
+        self.max_type_size
+    }
+
     fn memory64_enabled(&self) -> bool {
         self.memory64_enabled
     }
